#![cfg_attr(not(debug_assertions), windows_subsystem = "windows")] // hide console window on Windows in release

use eframe::{
    egui::{self, RichText, Slider, Window},
    epaint::Color32,
};
use egui_datepicker::{DatePicker, Local, NaiveDate};
use itertools::Itertools;
use serde::{Deserialize, Serialize};

fn main() -> Result<(), eframe::Error> {
    env_logger::init(); // Log to stderr (if you run with `RUST_LOG=debug`).
    let options = eframe::NativeOptions {
        initial_window_size: Some(egui::vec2(320.0, 240.0)),
        ..Default::default()
    };
    eframe::run_native(
        "Film Manager",
        options,
        Box::new(|cc| Box::new(MyApp::new(cc))),
    )
}

#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct Exposure {
    num: u8,
    #[serde(default)]
    printed: bool,
    #[serde(default)]
    damaged: bool,
    #[serde(default)]
    paper: Option<String>,
}

#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct Roll {
    id: String,
    name: String,
    #[serde(default)]
    desc: String,
    date: NaiveDate,
    exposures: Vec<Exposure>,
}

impl Roll {
    fn new(id: String, name: String, desc: String, date: NaiveDate, num_exposures: u8) -> Self {
        let mut exposures = Vec::new();
        for num in 1..=num_exposures {
            exposures.push(Exposure {
                num,
                printed: false,
                damaged: false,
                paper: None,
            })
        }
        Self {
            id,
            name,
            desc,
            date,
            exposures,
        }
    }

    pub fn short_str(&self) -> String {
        format!(
            "{}{} on {}, {} exp",
            self.id,
            if self.name.is_empty() {
                String::new()
            } else {
                format!(" - {}", self.name)
            },
            self.date.format("%m/%d/%Y"),
            self.exposures.len()
        )
    }
}

struct AppState {
    rolls: Vec<Roll>,
}

struct MyApp {
    state: AppState,
    roll_views: Vec<RollViewState>,
    new_roll_state: Option<NewRollState>,
}

struct RollViewState {
    roll: usize,
    confirm_delete: bool,
    exp: u8,
}

impl RollViewState {
    fn new(roll: usize) -> Self {
        Self {
            roll,
            confirm_delete: false,
            exp: 0,
        }
    }
}

struct NewRollState {
    id: String,
    name: String,
    desc: String,
    date: NaiveDate,
    exps: u8,
    show_id_req_err: bool,
}

impl Default for NewRollState {
    fn default() -> Self {
        Self {
            id: String::new(),
            name: String::new(),
            desc: String::new(),
            date: Local::now().date_naive(),
            exps: 24,
            show_id_req_err: false,
        }
    }
}

impl MyApp {
    fn new(cc: &eframe::CreationContext<'_>) -> Self {
        let rolls = eframe::get_value(cc.storage.unwrap(), "rolls").unwrap_or_else(|| {
            vec![Roll::new(
                "00001".into(),
                "Test".into(),
                "Testing".into(),
                Local::now().date_naive(),
                24,
            )]
        });
        Self {
            state: AppState { rolls },
            roll_views: Vec::new(),
            new_roll_state: None,
        }
    }
}

impl eframe::App for MyApp {
    fn save(&mut self, storage: &mut dyn eframe::Storage) {
        eframe::set_value(storage, "rolls", &self.state.rolls);
    }

    fn update(&mut self, ctx: &egui::Context, _frame: &mut eframe::Frame) {
        egui::CentralPanel::default().show(ctx, |ui| {
            ui.heading("Film Manager");
            ui.label("Rolls:");
            for (i, roll) in self.state.rolls.iter().enumerate() {
                if ui.button(roll.short_str()).clicked() && self.roll_views.iter().find(|v| v.roll == i).is_none() {
                    self.roll_views.push(RollViewState::new(i));
                }
            }
            if ui.button("Add").clicked() && self.new_roll_state.is_none() {
                self.new_roll_state = Some(NewRollState::default());
            }
            let mut to_close = Vec::new();
            for (i, roll_view_state) in self.roll_views.iter_mut().enumerate() {
                if roll_view_window(ctx, roll_view_state, &mut self.state) {
                    to_close.push(i);
                }
            }
            for i in to_close {
                self.roll_views.remove(i);
            }
            if let Some(new_roll_state) = self.new_roll_state.as_mut() {
                if new_roll_window(ctx, new_roll_state, &mut self.state) {
                    self.new_roll_state = None;
                }
            }
        });
    }
}

fn roll_view_window(
    ctx: &egui::Context,
    win_state: &mut RollViewState,
    app_state: &mut AppState,
) -> bool {
    let mut open = true;
    let window =
        Window::new(format!("Roll {}", app_state.rolls[win_state.roll].id)).open(&mut open);
    let deleted = window
        .show(ctx, |ui| {
            let roll = &mut app_state.rolls[win_state.roll];
            ui.label(format!("Name: {}", roll.name));
            ui.label(format!("Description: {}", roll.desc));
            ui.label(format!("Date: {}", roll.date.format("%Y-%m-%d")));
            ui.label("Exposures:");
            ui.horizontal_top(|ui| {
                ui.vertical(|ui| {
                    for (i, chunk) in roll.exposures.iter().chunks(5).into_iter().enumerate() {
                        ui.horizontal(|ui| {
                            for (j, exp) in chunk.enumerate() {
                                let num = i * 5 + j;
                                let mut label = RichText::new(format!("{:02}", exp.num));
                                if exp.printed {
                                    if exp.damaged {
                                        label = label.color(Color32::DARK_RED);
                                    } else {
                                        label = label.color(Color32::DARK_BLUE);
                                    }
                                }
                                ui.selectable_value(&mut win_state.exp, num as u8, label);
                            }
                        });
                    }
                });
                ui.vertical(|ui| {
                    let exp = &mut roll.exposures[win_state.exp as usize];
                    ui.checkbox(&mut exp.printed, "Printed");
                    if !exp.printed {
                        exp.damaged = false;
                    }
                    if exp.printed {
                        ui.checkbox(&mut exp.damaged, "Damaged");
                    }
                });
            });
            if ui.button("Delete").clicked() {
                win_state.confirm_delete = true;
            }
            if win_state.confirm_delete {
                Window::new("Confirm delete")
                    .show(ctx, |ui| {
                        ui.heading(format!(
                            "Really delete roll {}?",
                            app_state.rolls[win_state.roll].id
                        ));
                        if ui.button("No").clicked() {
                            win_state.confirm_delete = false;
                            return false;
                        }
                        if ui.button("Yes").clicked() {
                            app_state.rolls.remove(win_state.roll);
                            return true;
                        }
                        false
                    })
                    .unwrap()
                    .inner
                    .unwrap_or(false)
            } else {
                false
            }
        })
        .unwrap()
        .inner
        .unwrap_or(false);
    !open || deleted
}

fn new_roll_window(
    ctx: &egui::Context,
    win_state: &mut NewRollState,
    app_state: &mut AppState,
) -> bool {
    let mut open = true;
    Window::new("New Roll").open(&mut open).show(ctx, |ui| {
        ui.horizontal(|ui| {
            let label = ui.label("ID: ");
            ui.text_edit_singleline(&mut win_state.id)
                .labelled_by(label.id);
        });
        ui.horizontal(|ui| {
            let label = ui.label("Name: ");
            ui.text_edit_singleline(&mut win_state.name)
                .labelled_by(label.id);
        });
        let desc_label = ui.label("Description: ");
        ui.text_edit_multiline(&mut win_state.desc)
            .labelled_by(desc_label.id);
        ui.horizontal(|ui| {
            let label = ui.label("Date: ");
            ui.add(
                DatePicker::new("new_roll_date", &mut win_state.date)
                    .sunday_first(true)
                    .highlight_weekend(false),
            )
            .labelled_by(label.id);
        });
        ui.horizontal(|ui| {
            let label = ui.label("Num exposures: ");
            ui.add(Slider::new(&mut win_state.exps, 1..=50).step_by(1.0))
                .labelled_by(label.id);
        });
        if ui.button("Submit").clicked() {
            if win_state.id.is_empty() {
                win_state.show_id_req_err = true;
            } else {
                app_state.rolls.push(Roll::new(
                    win_state.id.clone(),
                    win_state.name.clone(),
                    win_state.desc.clone(),
                    win_state.date,
                    win_state.exps,
                ));
            }
        }
        if win_state.show_id_req_err {
            ui.colored_label(Color32::DARK_RED, "Error: ID required");
        }
    });
    !open
}
